package de.hft.stuttgart.citydoctor2.gui.tree;

import java.util.List;

import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.ErrorVisitor;
import de.hft.stuttgart.citydoctor2.check.error.AllPolygonsWrongOrientationError;
import de.hft.stuttgart.citydoctor2.check.error.AttributeInvalidError;
import de.hft.stuttgart.citydoctor2.check.error.AttributeMissingError;
import de.hft.stuttgart.citydoctor2.check.error.AttributeValueWrongError;
import de.hft.stuttgart.citydoctor2.check.error.ConsecutivePointSameError;
import de.hft.stuttgart.citydoctor2.check.error.DependenciesNotMetError;
import de.hft.stuttgart.citydoctor2.check.error.MultipleConnectedComponentsError;
import de.hft.stuttgart.citydoctor2.check.error.NestedRingError;
import de.hft.stuttgart.citydoctor2.check.error.NonManifoldEdgeError;
import de.hft.stuttgart.citydoctor2.check.error.NonManifoldVertexError;
import de.hft.stuttgart.citydoctor2.check.error.NonPlanarPolygonDistancePlaneError;
import de.hft.stuttgart.citydoctor2.check.error.NonPlanarPolygonNormalsDeviation;
import de.hft.stuttgart.citydoctor2.check.error.NotCeilingError;
import de.hft.stuttgart.citydoctor2.check.error.NotFloorError;
import de.hft.stuttgart.citydoctor2.check.error.NotGroundError;
import de.hft.stuttgart.citydoctor2.check.error.NotWallError;
import de.hft.stuttgart.citydoctor2.check.error.NullAreaError;
import de.hft.stuttgart.citydoctor2.check.error.PointTouchesEdgeError;
import de.hft.stuttgart.citydoctor2.check.error.PolygonHoleOutsideError;
import de.hft.stuttgart.citydoctor2.check.error.PolygonInteriorDisconnectedError;
import de.hft.stuttgart.citydoctor2.check.error.PolygonIntersectingRingsError;
import de.hft.stuttgart.citydoctor2.check.error.PolygonSameOrientationError;
import de.hft.stuttgart.citydoctor2.check.error.PolygonWithoutSurfaceError;
import de.hft.stuttgart.citydoctor2.check.error.PolygonWrongOrientationError;
import de.hft.stuttgart.citydoctor2.check.error.RingDuplicatePointError;
import de.hft.stuttgart.citydoctor2.check.error.RingEdgeIntersectionError;
import de.hft.stuttgart.citydoctor2.check.error.RingNotClosedError;
import de.hft.stuttgart.citydoctor2.check.error.RingTooFewPointsError;
import de.hft.stuttgart.citydoctor2.check.error.SchematronError;
import de.hft.stuttgart.citydoctor2.check.error.SolidNotClosedError;
import de.hft.stuttgart.citydoctor2.check.error.SolidSelfIntError;
import de.hft.stuttgart.citydoctor2.check.error.SurfaceUnfragmentedError;
import de.hft.stuttgart.citydoctor2.check.error.DegeneratedRingError;
import de.hft.stuttgart.citydoctor2.check.error.TooFewPolygonsError;
import de.hft.stuttgart.citydoctor2.check.error.UnknownCheckError;
import de.hft.stuttgart.citydoctor2.datastructure.ConcretePolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Edge;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.gui.CheckStatus;
import de.hft.stuttgart.citydoctor2.math.Triangle3d;
import de.hft.stuttgart.citydoctor2.utils.Localization;
import de.hft.stuttgart.citydoctor2.utils.PolygonIntersection;
import de.hft.stuttgart.citydoctor2.utils.PolygonIntersection.IntersectionType;
import javafx.scene.control.TreeItem;

public class ErrorItemVisitor implements ErrorVisitor {

	private static String nameOfAttribute = Localization.getText("ErrorItemVisitor.nameOfAttribute") + ": ";
	private static String childId = Localization.getText("ErrorItemVisitor.childId") + ": ";
	private final TreeItem<Renderable> root;

	public ErrorItemVisitor(TreeItem<Renderable> root) {
		this.root = root;
	}

	@Override
	public void visit(PolygonHoleOutsideError err) {
		PolygonNode polyNode = new PolygonNode(err.getPolygon(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> polyItem = new TreeItem<>(polyNode);
		root.getChildren().add(polyItem);
	}

	@Override
	public void visit(NonManifoldEdgeError err) {
		for (Edge e : err.getEdges()) {
			EdgeNode edgeNode = new EdgeNode(e);
			TreeItem<Renderable> edgeItem = new TreeItem<>(edgeNode);
			root.getChildren().add(edgeItem);
		}
	}

	@Override
	public void visit(MultipleConnectedComponentsError err) {
		for (int i = 0; i < err.getComponents().size(); i++) {
			TextNode textNode = new TextNode(Localization.getText("ErrorItemVisitor.component") + " " + i);
			TreeItem<Renderable> textItem = new TreeItem<>(textNode);
			root.getChildren().add(textItem);
			List<Polygon> component = err.getComponents().get(i);
			for (Polygon p : component) {
				PolygonNode polyNode = new PolygonNode(p, CheckStatus.NOT_CHECKED);
				TreeItem<Renderable> polyItem = new TreeItem<>(polyNode);
				textItem.getChildren().add(polyItem);
			}
		}
	}

	@Override
	public void visit(NestedRingError err) {
		PolygonNode polyNode = new PolygonNode(err.getPolygon(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> polyItem = new TreeItem<>(polyNode);
		root.getChildren().add(polyItem);
	}

	@Override
	public void visit(NonManifoldVertexError err) {
		VertexNode vertexNode = new VertexNode(err.getVertex());
		TreeItem<Renderable> vertexItem = new TreeItem<>(vertexNode);
		root.getChildren().add(vertexItem);
		for (int i = 0; i < err.getComponents().size(); i++) {
			TextNode textNode = new TextNode(Localization.getText("ErrorItemVisitor.component") + " " + i);
			TreeItem<Renderable> textItem = new TreeItem<>(textNode);
			root.getChildren().add(textItem);
			List<Polygon> component = err.getComponents().get(i);
			for (Polygon p : component) {
				PolygonNode polyNode = new PolygonNode(p, CheckStatus.NOT_CHECKED);
				TreeItem<Renderable> polyItem = new TreeItem<>(polyNode);
				textItem.getChildren().add(polyItem);
			}
		}
	}

	@Override
	public void visit(PolygonWrongOrientationError err) {
		for (Edge e : err.getEdges()) {
			EdgeNode edgeNode = new EdgeNode(e);
			TreeItem<Renderable> edgeItem = new TreeItem<>(edgeNode);
			root.getChildren().add(edgeItem);
		}
	}

	@Override
	public void visit(PolygonSameOrientationError err) {
		PolygonNode polyNode = new PolygonNode(err.getPolygon(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> polyItem = new TreeItem<>(polyNode);
		root.getChildren().add(polyItem);
	}

	@Override
	public void visit(SolidNotClosedError err) {
		for (Edge e : err.getErrorEdges()) {
			EdgeNode edgeNode = new EdgeNode(e);
			TreeItem<Renderable> edgeItem = new TreeItem<>(edgeNode);
			root.getChildren().add(edgeItem);
		}
	}

	@Override
	public void visit(DependenciesNotMetError err) {
		// not displayed
	}

	@Override
	public void visit(UnknownCheckError err) {
		// not displayed

	}

	@Override
	public void visit(RingNotClosedError err) {
		LinearRingNode ringNode = new LinearRingNode(err.getRing(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> ringItem = new TreeItem<>(ringNode);
		root.getChildren().add(ringItem);
	}

	@Override
	public void visit(ConsecutivePointSameError err) {
		LinearRingNode ringNode = new LinearRingNode(err.getRing(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> ringItem = new TreeItem<>(ringNode);
		root.getChildren().add(ringItem);

		VertexNode vertexNode = new VertexNode(err.getVertex1());
		TreeItem<Renderable> vertexItem = new TreeItem<>(vertexNode);
		root.getChildren().add(vertexItem);

		vertexNode = new VertexNode(err.getVertex2());
		vertexItem = new TreeItem<>(vertexNode);
		root.getChildren().add(vertexItem);
	}

	@Override
	public void visit(AllPolygonsWrongOrientationError err) {
		// not displayed
	}

	@Override
	public void visit(PolygonInteriorDisconnectedError err) {
		PolygonNode polyNode = new PolygonNode(err.getPolygon(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> polyItem = new TreeItem<>(polyNode);
		root.getChildren().add(polyItem);
	}

	@Override
	public void visit(NullAreaError err) {
		LinearRingNode ringNode = new LinearRingNode(err.getRing(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> ringItem = new TreeItem<>(ringNode);
		root.getChildren().add(ringItem);
	}

	@Override
	public void visit(RingTooFewPointsError err) {
		LinearRingNode ringNode = new LinearRingNode(err.getRing(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> ringItem = new TreeItem<>(ringNode);
		root.getChildren().add(ringItem);

		if (err.getRing().getGmlId().isGenerated()) {
			PolygonNode polyNode = new PolygonNode(err.getRing().getParent(), CheckStatus.NOT_CHECKED);
			TreeItem<Renderable> polyItem = new TreeItem<>(polyNode);
			ringItem.getChildren().add(polyItem);
		}
	}

	@Override
	public void visit(NonPlanarPolygonNormalsDeviation err) {
		PolygonNode polyNode = new PolygonNode(err.getPolygon(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> polyItem = new TreeItem<>(polyNode);
		root.getChildren().add(polyItem);

		TextNode textNode = new TextNode(
				Localization.getText("ErrorItemVisitor.deviation") + ": " + err.getDeviation());
		TreeItem<Renderable> textItem = new TreeItem<>(textNode);
		root.getChildren().add(textItem);
	}

	@Override
	public void visit(NonPlanarPolygonDistancePlaneError err) {
		PolygonNode polyNode = new PolygonNode(err.getPolygon(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> polyItem = new TreeItem<>(polyNode);
		root.getChildren().add(polyItem);

		VertexNode vertexNode = new VertexNode(err.getVertex());
		TreeItem<Renderable> vertexItem = new TreeItem<>(vertexNode);
		root.getChildren().add(vertexItem);

		TextNode textNode = new TextNode(
				Localization.getText("ErrorItemVisitor.distance") + ": " + err.getDistance() + "m");
		TreeItem<Renderable> textItem = new TreeItem<>(textNode);
		root.getChildren().add(textItem);

	}

	@Override
	public void visit(PolygonIntersectingRingsError err) {
		PolygonNode polyNode = new PolygonNode(err.getPolygon(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> polyItem = new TreeItem<>(polyNode);
		root.getChildren().add(polyItem);
	}

	@Override
	public void visit(SolidSelfIntError err) {
		List<PolygonIntersection> intersections = err.getIntersections();
		for (int i = 0; i < intersections.size(); i++) {
			PolygonIntersection pi = intersections.get(i);
			TextNode intersectionNode = new TextNode(Localization.getText("ErrorItemVisitor.intersection") + " " + i);
			TreeItem<Renderable> intersectionItem = new TreeItem<>(intersectionNode);
			root.getChildren().add(intersectionItem);

			PolygonNode polyNode1 = new PolygonNode(pi.getP1(), CheckStatus.NOT_CHECKED);
			TreeItem<Renderable> polyItem1 = new TreeItem<>(polyNode1);
			intersectionItem.getChildren().add(polyItem1);
			PolygonNode polyNode2 = new PolygonNode(pi.getP2(), CheckStatus.NOT_CHECKED);
			TreeItem<Renderable> polyItem2 = new TreeItem<>(polyNode2);
			intersectionItem.getChildren().add(polyItem2);
			PolygonIntersection intersection = err.getIntersections().get(0);
			if (intersection.getType() == IntersectionType.TRIANGLES) {
				TextNode trianglesNode = new TextNode(Localization.getText("ErrorItemVisitor.triangles"));
				TreeItem<Renderable> trianglesItem = new TreeItem<>(trianglesNode);
				intersectionItem.getChildren().add(trianglesItem);

				ConcretePolygon c = createPolygonForTriangle(intersection.getT1());
				PolygonNode pN1 = new PolygonNode(c, CheckStatus.NOT_CHECKED);
				TreeItem<Renderable> pI1 = new TreeItem<>(pN1);
				trianglesItem.getChildren().add(pI1);

				ConcretePolygon c2 = createPolygonForTriangle(intersection.getT2());
				PolygonNode pN2 = new PolygonNode(c2, CheckStatus.NOT_CHECKED);
				TreeItem<Renderable> pI2 = new TreeItem<>(pN2);
				trianglesItem.getChildren().add(pI2);
			}
		}
	}

	private ConcretePolygon createPolygonForTriangle(Triangle3d t) {
		ConcretePolygon c = new ConcretePolygon();
		LinearRing extRing = new LinearRing(LinearRingType.EXTERIOR);
		Vertex start = new Vertex(t.getP1());
		extRing.getVertices().add(start);
		extRing.getVertices().add(new Vertex(t.getP2()));
		extRing.getVertices().add(new Vertex(t.getP3()));
		extRing.getVertices().add(start);
		c.setExteriorRing(extRing);
		return c;
	}

	@Override
	public void visit(TooFewPolygonsError err) {
		// not displayed

	}

	@Override
	public void visit(RingDuplicatePointError err) {
		LinearRingNode ringNode = new LinearRingNode(err.getRing(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> ringItem = new TreeItem<>(ringNode);
		root.getChildren().add(ringItem);

		VertexNode vertexNode = new VertexNode(err.getVertex1());
		TreeItem<Renderable> vertexItem = new TreeItem<>(vertexNode);
		root.getChildren().add(vertexItem);

		vertexNode = new VertexNode(err.getVertex2());
		vertexItem = new TreeItem<>(vertexNode);
		root.getChildren().add(vertexItem);

	}

	@Override
	public void visit(RingEdgeIntersectionError err) {
		LinearRingNode ringNode = new LinearRingNode(err.getRing(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> ringItem = new TreeItem<>(ringNode);
		root.getChildren().add(ringItem);

		EdgeNode edgeNode = new EdgeNode(err.getEdge1());
		TreeItem<Renderable> edgeItem = new TreeItem<>(edgeNode);
		root.getChildren().add(edgeItem);

		EdgeNode edge2Node = new EdgeNode(err.getEdge2());
		TreeItem<Renderable> edge2Item = new TreeItem<>(edge2Node);
		root.getChildren().add(edge2Item);

		VertexNode vertexNode = new VertexNode(new Vertex(err.getIntersection()));
		TreeItem<Renderable> vertexItem = new TreeItem<>(vertexNode);
		root.getChildren().add(vertexItem);
	}

	@Override
	public void visit(PointTouchesEdgeError err) {
		TextNode textNode = new TextNode(Localization.getText("ErrorItemVisitor.pointTouchesEdge"));
		TreeItem<Renderable> textItem = new TreeItem<>(textNode);
		root.getChildren().add(textItem);

		VertexNode vertexNode = new VertexNode(err.getVertex());
		TreeItem<Renderable> vertexItem = new TreeItem<>(vertexNode);
		root.getChildren().add(vertexItem);

		EdgeNode edgeNode = new EdgeNode(err.getEdge());
		TreeItem<Renderable> edgeItem = new TreeItem<>(edgeNode);
		root.getChildren().add(edgeItem);

		LinearRingNode ringNode = new LinearRingNode(err.getRing(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> ringItem = new TreeItem<>(ringNode);
		root.getChildren().add(ringItem);
	}

	@Override
	public void visit(NotCeilingError err) {
		PolygonNode polyNode = new PolygonNode(err.getPolygon(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> polyItem = new TreeItem<>(polyNode);
		root.getChildren().add(polyItem);
	}

	@Override
	public void visit(NotFloorError err) {
		PolygonNode polyNode = new PolygonNode(err.getPolygon(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> polyItem = new TreeItem<>(polyNode);
		root.getChildren().add(polyItem);
	}

	@Override
	public void visit(NotWallError err) {
		PolygonNode polyNode = new PolygonNode(err.getPolygon(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> polyItem = new TreeItem<>(polyNode);
		root.getChildren().add(polyItem);
	}

	@Override
	public void visit(NotGroundError err) {
		PolygonNode polyNode = new PolygonNode(err.getPolygon(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> polyItem = new TreeItem<>(polyNode);
		root.getChildren().add(polyItem);
	}

	@Override
	public void visit(CheckError err) {
		// not used
	}

	@Override
	public void visit(SchematronError err) {
		TextNode textNode = new TextNode(err.getErrorIdString());
		TreeItem<Renderable> textItem = new TreeItem<>(textNode);
		root.getChildren().add(textItem);
	}

	@Override
	public void visit(SurfaceUnfragmentedError err) {
		TextNode textNode = new TextNode(
				Localization.getText("ErrorItemVisitor.deviation") + ": " + err.getAngleDeviation());
		TreeItem<Renderable> textItem = new TreeItem<>(textNode);
		root.getChildren().add(textItem);
	}

	@Override
	public void visit(DegeneratedRingError err) {
		TextNode textNode = new TextNode(Localization.getText("ErrorItemVisitor.degeneratedRing"));
		TreeItem<Renderable> textItem = new TreeItem<>(textNode);
		root.getChildren().add(textItem);

		LinearRingNode polyNode = new LinearRingNode(err.getRing(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> polyItem = new TreeItem<>(polyNode);
		root.getChildren().add(polyItem);
	}

	@Override
	public void visit(AttributeMissingError err) {
		if (!err.getChildId().isEmpty()) {
			TextNode textNode = new TextNode(childId + err.getChildId());
			TreeItem<Renderable> textItem = new TreeItem<>(textNode);
			root.getChildren().add(textItem);
		}
		TextNode nameNode = new TextNode(nameOfAttribute + err.getNameOfAttribute());
		TreeItem<Renderable> nameItem = new TreeItem<>(nameNode);
		root.getChildren().add(nameItem);

	}

	@Override
	public void visit(AttributeValueWrongError err) {
		if (!err.getChildId().isEmpty()) {
			TextNode textNode = new TextNode(childId + err.getChildId());
			TreeItem<Renderable> textItem = new TreeItem<>(textNode);
			root.getChildren().add(textItem);
		}
		TextNode nameNode = new TextNode(nameOfAttribute + err.getNameOfAttribute());
		TreeItem<Renderable> nameItem = new TreeItem<>(nameNode);
		root.getChildren().add(nameItem);
	}

	@Override
	public void visit(AttributeInvalidError err) {
		if (!err.getChildId().isEmpty()) {
			TextNode textNode = new TextNode(childId + err.getChildId());
			TreeItem<Renderable> textItem = new TreeItem<>(textNode);
			root.getChildren().add(textItem);
		}
		TextNode nameNode = new TextNode(nameOfAttribute + err.getNameOfAttribute());
		TreeItem<Renderable> nameItem = new TreeItem<>(nameNode);
		root.getChildren().add(nameItem);
	}

	@Override
	public void visit(PolygonWithoutSurfaceError err) {
		PolygonNode polyNode = new PolygonNode(err.getPolygon(), CheckStatus.NOT_CHECKED);
		TreeItem<Renderable> polyItem = new TreeItem<>(polyNode);
		root.getChildren().add(polyItem);
	}

}
